/*
 * ioGame
 * Copyright (C) 2021 - 2023  渔民小镇 （262610965@qq.com、luoyizhu@gmail.com） . All Rights Reserved.
 * # iohao.com . 渔民小镇
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.iohao.mmo.common.processor.hook;

import com.iohao.game.action.skeleton.core.BarSkeleton;
import com.iohao.game.action.skeleton.core.flow.FlowContext;
import com.iohao.game.action.skeleton.protocol.HeadMetadata;
import com.iohao.game.bolt.broker.core.common.processor.hook.RequestMessageClientProcessorHook;
import com.iohao.game.common.kit.ExecutorKit;
import com.iohao.game.common.kit.concurrent.ThreadCreator;
import com.iohao.mmo.common.annotation.IoThread;

import java.lang.reflect.Method;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadFactory;

/**
 * @author 渔民小镇
 * @date 2023-07-24
 */
public final class MyRequestMessageClientProcessorHook implements RequestMessageClientProcessorHook {
    final Executor[] executors;
    final int executorLength;
    final ExecutorService cacheThreadPool = ExecutorKit
            .newCacheThreadPool("ioGameMMO");

    public MyRequestMessageClientProcessorHook() {
        int availableProcessors = availableProcessors2n();
        executorLength = availableProcessors;
        executors = new Executor[executorLength];

        for (int i = 0; i < availableProcessors; i++) {
            // 线程名：RequestMessage-线程总数-当前线程编号
            int threadNo = i + 1;
            String threadNamePrefix = String.format("RequestMessage-%s-%s", availableProcessors, threadNo);
            executors[i] = ExecutorKit.newSingleThreadExecutor(new TheThreadFactory(threadNamePrefix));
        }
    }

    @Override
    public void processLogic(BarSkeleton barSkeleton, FlowContext flowContext) {

        HeadMetadata headMetadata = flowContext.getRequest().getHeadMetadata();
        Method actionMethod = flowContext.getActionCommand().getActionMethod();
        var annotation = actionMethod.getAnnotation(IoThread.class);
        if (Objects.nonNull(annotation)) {
            cacheThreadPool.execute(() -> barSkeleton.handle(flowContext));
            return;
        }

        long userId = flowContext.getUserId();

        if (userId == 0) {
            // 如果没有登录，使用 channelId 计算；如果 channelId 不存在，则使用 cmd
            userId = Optional.ofNullable(headMetadata.getChannelId())
                    .map(String::hashCode)
                    .map(Math::abs)
                    .orElseGet(headMetadata::getCmdMerge);
        }

        // 根据 userId 获取对应的 Executor
        int index = (int) (userId & (executorLength - 1));

        // 执行业务框架
        executors[index].execute(() -> barSkeleton.handle(flowContext));
    }

    private int availableProcessors2n() {
        int n = Runtime.getRuntime().availableProcessors();
        n |= (n >> 1);
        n |= (n >> 2);
        n |= (n >> 4);
        n |= (n >> 8);
        n |= (n >> 16);
        return (n + 1) >> 1;
    }

    private static class TheThreadFactory extends ThreadCreator implements ThreadFactory {

        public TheThreadFactory(String threadNamePrefix) {
            super(threadNamePrefix);
            this.setDaemon(true);
        }

        @Override
        public Thread newThread(Runnable runnable) {
            return createThread(runnable);
        }

        @Override
        protected String nextThreadName() {
            return this.getThreadNamePrefix();
        }
    }
}
